/*
* $Log: tokens.c,v $
* Revision 1.5  2004/01/30 23:10:52  Default
* Dynamic memory stuff removed - replaced by static ewram assignment
* Bug fixed the "He stands" string for the case when the lord
* has an army or is looking at one.
*
* Revision 1.4  2004/01/26 20:41:49  richq
* Added a way to append the area name to the token buffer.
* Plus some tidying up and bios div instead of %.
*
* Revision 1.3  2004/01/22 20:32:49  richq
* Bug fix for the MildlyCold code in .c, .h has a function name corrected
* and the current graphics row made "public".
*
* Revision 1.2  2004/01/22 19:53:34  richq
* Helper functions added (addUppercaseToken, etc).
* Number to words added.
* Removed the need to reset the token list manually, now done when
* the list is flushed.
*
* Revision 1.1  2004/01/21 19:33:09  Default
* Token parsing - tokens are stored using 5 bit bytes compressed together,
* the format is [length][length * char] - code here expands the tokens,
* prints them out and other tokeny things.
*
*/

/*
* Everything to do with text tokens and token parsing
* See tokendata for definitions
*/

#include "keypad.h"
#include "qmath.h"

#include "tokens.h"
#include "tokendata.h"
#include "location.h"
#include "gfx.h"
#include "stdlib.h"

#include "memory_defs.h"

#define TOKEN_LIST_SIZE 0x100
#define TOKEN_BUFFER_SIZE 0x20
#define FINAL_ROW 20


// define the tokens list, 
// the list contains the tokens (8-bit values)
unsigned char * const g_tokenList = (char*)ADDR_G_TOKENLIST;
unsigned char * g_heap;       // current position in g_tokenList

// the buffer contains the ASCII characters
char * const g_tokenBuffer = (char*)ADDR_G_TOKENBUFFER;
char * g_bufferPos;           // current position in token buffer

// the current size of the word/token being expanded
int g_tokenLength;
// uppercase capitalises first letter, capslock capitalises whole sentences
int g_uppercase;
int g_capsLock;

// pluralness is usually based on the previous numberToWords value
int g_isPlural;

// used for text display:
int g_startX;
int g_startY;
int g_indent;   // where to start X when a newline is reached (left hand side)
int g_maxX;     // the end of the line (right hand side)
int g_currentX;
int g_currentY;


void init_token_buffers(void) {
  g_heap = g_tokenList;
}


// returns the i-th five bit value from the packed 5 bit value array
int fiveBitAt(int i) {
  int shift = (i*5)&7;
  int index = (i*5)/8;
  int data = (tokendict[index])|(tokendict[index+1]<<8);
  if (shift > 0)
    data >>= shift;
  data &= 0x1f;
  return data;
}


// places ASCII values for token "tok" into g_bufferPos and updates g_tokenLength
// make sure g_bufferPos is valid before calling!!!!
void getToken(int tok) {
  int index = 0;
  int i, j;
  if (tok < 0 || tok >= NEWLINE)
    return;
  
  for (i = 0; i < tok; i++) {
    
    int size = fiveBitAt(index);
    index += (size);
  }
  int size = fiveBitAt(index);
  size--;
  index++;
  g_tokenLength += size;
  for (j = 0; j < size; j++) {
    int data = fiveBitAt(index);
    char theChar = (char)((data|0x60)&0xff);
    *g_bufferPos++ = theChar;
    index++;
  }
}

u8* addLiteral(char literal, u8 *heap) {
  *heap++ = LITERAL;
  *heap++ = literal;
  return heap;
}

// add the name of an area to heap (e.g. the Moon, Blood)
u8 * addArea(int area, u8 * heap) {
  switch (area) {
    case 0x07:
    case 0x20:
    case 0x39:
              *heap++ = 0; //the
              /* NO BREAK */
    default : 
              *heap++ = UPPERCASE;
              *heap++ = area;
  }
  return heap;
}


void addAreaToken(int area) {
  g_heap = addArea(area, g_heap);
}

void addLiteralToken(char lit) {
  *g_heap++ = LITERAL;
  *g_heap++ = lit;
}

void addNewlineToken(void) {
  *g_heap++ = NEWLINE;
}


void addUppercaseToken(void) {
  *g_heap++ = UPPERCASE;
}

// pluralise if needed
void addPluralisedToken(u8 t) {
  *g_heap++ = t;
  if (g_isPlural)
    addLiteralToken('s');
}

void addSingleToken(u8 t) {
  *g_heap++ = t;
}




// copy a given token array to the heap
// copies until *from == ENDTOKEN
u8* addTokenArray(const u8 * from, u8 * to) 
{
  while (*from != ENDTOKEN)
    *to++ = *from++;
  return to;
  // *heap = ENDTOKEN; //?
}

void addToTokenList(const u8 * array) {
  g_heap = addTokenArray(array, g_heap);
}

int isTerrainPlural(int terrain) {
  return (terrain == MOUNTAIN ||
  terrain == DOWNS ||
  terrain == WASTES ||
  terrain == ARMY ||      // we pretend army is plains anyway
  terrain == PLAIN);
}


// adds the Domain string to the heap
u8 * createDomain(locdat_t * loc, u8 * heap) {
  
  if (loc->domain == 0) {
    if (loc->terrain == HENGE) {
      *heap++ = UPPERCASE;
      *heap++ = loc->area;
      *heap++ = NOSPACE;
      *heap++ = 0x44;
    } else
    if (loc->terrain == LAKE) {
      heap = addTokenArray(lake, heap);
      *heap++ = loc->area;
    } else
    if (loc->terrain == WASTES) {
      heap = addTokenArray(wastes, heap);
      
    } else {
      *heap++ = 0;      // the
      *heap++ = UPPERCASE;
      if (loc->terrain == ARMY)     // pretend it's plains
        *heap++ = loc->terrain+0x42;
      else
        *heap++ = loc->terrain+0x41;
      if ( isTerrainPlural(loc->terrain)
      ) {
        heap = addLiteral('s', heap);
//        *heap++ = LITERAL;
//        *heap++ = 's';
      }  
      
      *heap++ = 0x40;  // of
      heap = addArea(loc->area, heap);
    }
    
      
  } else { // domain != 0
    if (!isTerrainPlural(loc->terrain)) {
      *heap++ = 0x53;   // a
    }
    
    if (loc->terrain == ARMY)
      *heap++ = loc->terrain+0x42;
    else
      *heap++ = loc->terrain+0x41;
    
    if ( isTerrainPlural(loc->terrain) ) {
      heap = addLiteral('s', heap);
//      *heap++ = LITERAL;
//      *heap++ = 's';
    }
    
    heap = addTokenArray(indomof, heap);
    heap = addArea(loc->area, heap); 
  }
  return heap;
}

void addDomainTokens(locdat_t * loc) {
  g_heap = createDomain(loc, g_heap);
}


//void reset_heap(void) {
//  g_heap = tokenList;
//}


void addNameToken(int name) {
  *g_heap++ = UPPERCASE;
  *g_heap++ = name;
}

void addNameThePlusToken(int name, int t) {
  addNameToken(name);
  *g_heap++ = 0;
  *g_heap++ = UPPERCASE;
  *g_heap++ = t;
}
u8* addTokenNospace(int t, u8 * heap) {
  *heap++ = NOSPACE;
  *heap++ = t;
  return heap;
}


void createLordTitle(int title, int name) {
   
  // 0x00 FirstName "the Moonprince"
  // 0x01 FirstName
  // 0x02 FirstName "the Fey"
  // 0x03 FirstName "the Wise"
  // 0x04 FirstName "the Dragonlord"
  // 0x05 FirstName "the Skulkrin"
  // less than 0x0A "The Lord of" FirstName
  // else "Lord" FirstName
  
  if (title == 0x00) {
    addNameThePlusToken(name, 0x7e);
  } else
  if (title == 0x01) {
    addNameToken(name);
  } else
  if (title == 0x02) {
    addNameThePlusToken(name, 0x7a);
  } else
  if (title == 0x03) {
    addNameThePlusToken(name, 0x7b);
  } else
  if (title == 0x04) {
    addNameThePlusToken(name, 0x7d);  //dragon
    g_heap = addTokenNospace(0x79,g_heap);      //lord
  } else
  if (title == 0x05) {
    addNameThePlusToken(name, 0x7c);
  } else
  if (title < 0x0a) {
    g_heap = addTokenArray(theLordOf, g_heap);
    addNameToken(name);

  } else
  if (title == 0x0a) {
    *g_heap++ = 0;          // the
    addNameToken(name);

    *g_heap++ = 0x40;       // of
    addNameToken(name);

  } else {
    g_heap = addTokenArray(Lord, g_heap);
    addNameToken(name);

  }
  
}

// called in think...
void createLordTitleUppercase1(int title, int name) {
//  reset_heap();
  *g_heap++ = UPPERCASE;
  
  createLordTitle(title, name);
}

//f94d
void addDirection(int direction) {
  *g_heap++ = directions[direction][0];
  if (directions[direction][1] != 0) {
    g_heap = addTokenNospace(directions[direction][1],g_heap);
  }
}


// create the string "he stands at blah looking north to the foo of bar"
void createHeStandsString(location_t * position) {
  
  g_heap = g_tokenList;
  g_heap = addTokenArray(heStands, g_heap);
  g_capsLock = 0;
  locdat_t loc;
  loc.x = position->x;
  loc.y = position->y;
  fillLocationData(&loc);
  
  *g_heap++ = phrasalEndings[loc.terrain];
  
  g_heap = createDomain(&loc, g_heap);
  
  g_heap = addLiteral(',', g_heap);
  
  *g_heap++ = 0x5b;   // looking
  *g_heap++ = UPPERCASE;
  addDirection(position->direction);
  
  
  // use f9d3 now to get the thing we look at here.
  locdat_t lookAtLoc;
  lookAtLoc.x = lookAtPos.x;
  lookAtLoc.y = lookAtPos.y;
  fillLocationData(&lookAtLoc);
  
  // bug fix - you get "He stands on the plains looking east to the plains"
  // if looking at an army or lord himself has an army.
  if (loc.terrain == ARMY) 
    loc.terrain = PLAIN;
  if (lookAtLoc.terrain == ARMY) 
    lookAtLoc.terrain = PLAIN;
  
  if (loc.area != lookAtLoc.area || 
      loc.domain != lookAtLoc.domain || 
      loc.terrain != lookAtLoc.terrain) {
    *g_heap++ = 0x5c; // to
    g_heap = createDomain(&lookAtLoc, g_heap);
  }
  
  g_heap = addLiteral('.', g_heap);

  
}


void makeUpperCase(char * theChar) {
  unsigned char c = (unsigned char)(*theChar);
  if (!g_capsLock) {
    if (!g_uppercase) {
      return;
    }
    g_uppercase = 0;
  }
  // got here? must make upper case.
  if (c < (unsigned char)'a' || c > (unsigned char)'z')
  {
    return;
  }
  *theChar &= 0xdf;
}


// f6c0
void print_current_buffer(int palette) {
  int finalx = g_currentX + g_tokenLength;
  if (finalx >= g_maxX) {
    g_currentX = g_indent-1;
    g_currentY++;
  }
  g_currentX++;
  if (g_currentY == FINAL_ROW) { 
    // end of screen
    // wait for A press, clear text, reset position then carry on
    WaitKeyPress(KEY_A);
    clear_text_from(g_startY);
    g_currentY = g_startY;
    g_currentX = g_startX;
    // return;
  }
  // ok to draw text...
  // use void print_text(char text, int row, int col, int pal) in gfx.
  g_bufferPos = g_tokenBuffer;
  char theChar;
  do {
    theChar = *g_bufferPos++;
    makeUpperCase(&theChar);
    print_text(theChar, g_currentX, g_currentY, palette);
    g_currentX++;
  }
  while (--g_tokenLength != 0);
  
}


// only call during token expansion
void addASCIIchar(void) {
  //f768  
  g_heap++; //skip literal token
  *g_bufferPos++ = *g_heap++;
  g_tokenLength++;
}

void tokenAndEnd(void) {
  getToken(*g_heap);
  g_heap++;
  while (*g_heap == LITERAL)  {
    addASCIIchar();
  }
  if (*g_heap == NOSPACE)  {
    g_heap++;
    tokenAndEnd();
  }
}

// f7b5
void print_tokens(int y, int x, int indent, int maxx, int palette) {
  
  g_startX = x;
  g_startY = y;
  g_indent = indent;
  g_maxX = maxx;
  g_currentX = g_startX-1; 
  g_currentY = g_startY; 
  
  print_tokens_pal(palette);
}

void print_tokens_pal(int palette) {
  
  *g_heap++ = ENDTOKEN;
  
  // f717
  g_heap = g_tokenList;
  g_bufferPos = g_tokenBuffer;
  
  g_uppercase = 0;
  g_tokenLength = 0;
  
  while (*g_heap != ENDTOKEN) {
    if (*g_heap == LITERAL)  {
      addASCIIchar();
      // jump to after decompress token
      while (*g_heap == LITERAL)  {
        addASCIIchar();
      }
      if (*g_heap == NOSPACE)  {
        g_heap++;
        tokenAndEnd();
      }
    } else if (*g_heap == UPPERCASE)  {
      g_heap++;
      g_uppercase = 1;
      // restart loop
      continue;
    } else if (*g_heap == NEWLINE)  {
      g_heap++;
      g_currentX = g_indent; 
      g_currentY++;
      // restart loop
      continue;
    } else {
      tokenAndEnd();
    }
    
    // f74a - finished decompressing, draw the text!
    print_current_buffer(palette);
    g_bufferPos = g_tokenBuffer;
    
  }  
  g_heap = g_tokenList;
}


//void divMod10(int top, int * result, int * modulus) {
//  // replace with gbadiv
//  *result = top/10;
//  *modulus = top%10;  
//}

void setTens(int largeVal,
             int * units,
             int * tens,
             int * hundreds,
             int * thousands,
             int * tensthous
             ) {
  
  signed long result, modulus;
  FastIntDivide(largeVal, 10, &result, &modulus);
  *units = modulus;
  FastIntDivide(result, 10, &result, &modulus);

  *tens = modulus;
  FastIntDivide(result, 10, &result, &modulus);

  *hundreds = modulus;
  FastIntDivide(result, 10, &result, &modulus);

  *thousands = modulus;
  FastIntDivide(result, 10, &result, &modulus);

  *tensthous = modulus;
  
}



u8* addTeenEnding(int teen, u8 *heap) {
  if (teen > 2) {
    if (teen == 8) {
      heap = addTokenArray(een, heap);
    } else {
      heap = addTokenNospace(0xb8, heap);
    }
  }
  return heap; 
}


// code at e1aa
u8 *  addNumber(int tens, int units, u8 * heap) {
  if (tens == 0) {
    // one to ten
    *heap++ = units+0x62;
  } else if (tens == 1) {
    // e1eb
    // add the teens, 10-19
    *heap++ = theTeens[units];
    heap = addTeenEnding(units, heap);
  } else {
    *heap++ = theTees[tens-2];
    if (tens > 2) {
      if (tens != 8) {
        heap = addLiteral('t', heap);
      } 
      heap = addLiteral('y', heap);
    }
    
    if (units != 0) {
      heap = addTokenArray(dashJoin, heap);
      *heap++ = units+0x62;
    }
  }
  return heap;  
}


/*
  based on e21f
  Takes an int value and converts it to words 
  eg. 2001 -> "two thousand and one"
  This routine adds the tokens for the words to the token list
*/
void numberToWords(int value) {
  
  
  //int index = 0;
  int units, tens, hundreds, thousands, tensthous;
  setTens(value, &units, &tens, &hundreds, &thousands, &tensthous);
  int total = units + tens + hundreds + thousands + tensthous;
  if (total == 0) {
    *g_heap++ = 0xbc;     // no
    return;
  }
  
  // check plural too...
  g_isPlural = 1;
  if (total == 1 && units == 1)
    g_isPlural = 0;
  int tensAndUnits = tens+units;
  int hundredsAndThousands = hundreds + thousands + tensthous;
  if (hundredsAndThousands != 0) {
    int justThousands = thousands + tensthous;
    if (justThousands != 0) {
      // e25b
      g_heap = addNumber(tensthous, thousands, g_heap);
      *g_heap++ = 0xbb; // thousand
    }
    if (hundreds != 0) {
      g_heap = addNumber(0, hundreds, g_heap);
      *g_heap++ = 0xba; // hundred
    }
    if (tensAndUnits  != 0) {
      *g_heap++ = 0x87; // and
    }  
  }
  
  if (tensAndUnits != 0) {
    g_heap = addNumber(tens, units, g_heap);
  }
  
  //*g_heap++ = 0xff; 
}


/*
 Adds strings such as "utterly bold" to the token list.
 quantity       description
    0         utterly     negative
    1         very        negative
    2                     negative
    3         quite       negative
    4         sligtly     negative
    5                     positive
    6         very        positive
    7         utterly     positive
*/
void addMildlyCold(int quantity, int positive, int negative) {
  
  int noun = positive; // be optimistic
  if (quantity < 5) {
    noun = negative;
  }
  
  int adjective = 0xd1; //utterly
  
  switch (quantity) {
                                          //if (quantity == 3)  
    case 3: adjective++; //quite (+3)     //adjective+=3;
          /*No break*/                    //add adjective, noun
    
    case 4: adjective++; //slightly (+2)  //if (quantity == 4)
          /*No break*/                    //adjective+=2; //add adjective, noun
    case 6:                               //if (quantity == 6 ||quantity == 1) 
    case 1: adjective++; //very (+1)      //adjective+=1;
          /*No break*/                    //add adjective, noun
    case 0:                               //if (quantity == 0 ||quantity == 7)  
    case 7: *g_heap++ = adjective;        //add adjective, noun
    default : break;                      // if (quantity == 2 || quantity == 5)  
                                          // add noun only 
  }
  
  *g_heap++ = noun;
}

